@@ -25,6 +25,7 @@
 #include <unistd.h>
 #include <string.h> 
 #include <glib.h>
+#include <glib/gstdio.h>
 #include <glib-object.h>
 #include <gnokii.h>
 
@@ -41,6 +42,8 @@ static GHashTable *driver_device = NULL;
  * match */
 static GHashTable *driver_model = NULL;
 
+static void phonemgr_utils_init_hash_tables (void);
+
 void
 phonemgr_utils_gn_statemachine_clear (struct gn_statemachine *state)
 {
@@ -101,10 +104,20 @@ phonemgr_utils_gn_error_to_string (gn_error error, PhoneMgrError *perr)
 	}
 }
 
-gboolean
-phonemgr_utils_is_bluetooth (const char *addr)
+PhonemgrConnectionType
+phonemgr_utils_address_is (const char *addr)
 {
-	return (g_file_test (addr, G_FILE_TEST_EXISTS) == FALSE);
+	struct stat buf;
+
+	if (g_stat (addr, &buf) > 0 && S_ISCHR (buf.st_mode)) {
+		//FIXME this could also be IrDA
+		return PHONEMGR_CONNECTION_SERIAL;
+	} else if (bachk (addr) == 0) {
+		return PHONEMGR_CONNECTION_BLUETOOTH;
+	} else {
+		//FIXME find a better match
+		return PHONEMGR_CONNECTION_USB;
+	}
 }
 
 static int
@@ -112,48 +125,60 @@ get_rfcomm_channel (sdp_record_t *rec)
 {
 	int channel = -1;
 	sdp_list_t *protos = NULL;
+	sdp_data_t *data;
+	char *name = NULL;
 
 	if (sdp_get_access_protos (rec, &protos) != 0)
 		goto end;
 
+	data = sdp_data_get(rec, SDP_ATTR_SVCNAME_PRIMARY);
+	if (data)
+		name = g_strdup_printf ("%.*s", data->unitSize, data->val.str);
+
+	/* We can't seem to connect to the PC Suite channel */
+	if (strstr (name, "Nokia PC Suite") != NULL)
+		goto end;
+	/* And that type of channel on Nokia Symbian phones doesn't
+	 * work either */
+	if (strstr (name, "Bluetooth Serial Port") != NULL)
+		goto end;
+	/* Avoid the m-Router channel, same as the PC Suite on Sony Ericsson phones */
+	if (strstr (name, "m-Router Connectivity") != NULL)
+		goto end;
+
 	channel = sdp_get_proto_port (protos, RFCOMM_UUID);
- end:
+end:
+	g_free (name);
 	sdp_list_foreach(protos, (sdp_list_func_t)sdp_list_free, 0);
 	sdp_list_free(protos, 0);
 	return channel;
 }
 
-/* Determine whether the given device supports Dial-Up Networking, and if so
+/* Determine whether the given device supports Serial or Dial-Up Networking, and if so
  * what the RFCOMM channel number for the service is.
  */
 static int
-find_dun_channel (bdaddr_t *adapter, bdaddr_t *device)
+find_service_channel (bdaddr_t *adapter, bdaddr_t *device, uint16_t svclass_id)
 {
 	sdp_session_t *sdp = NULL;
 	sdp_list_t *search = NULL, *attrs = NULL, *recs = NULL, *tmp;
-	uuid_t browse_uuid, dun_id, obex_uuid;
-	uint16_t proto_desc_list = SDP_ATTR_PROTO_DESC_LIST;
-	uint16_t svclass_id_list = SDP_ATTR_SVCLASS_ID_LIST;
+	uuid_t browse_uuid, service_id, obex_uuid;
+	uint32_t range = 0x0000ffff;
 	int channel = -1;
 
 	sdp = sdp_connect (adapter, device, SDP_RETRY_IF_BUSY);
 	if (!sdp)
 		goto end;
 
-	/* Normally, we'd just do a search for OBEX_FILETRANS_SVCLASS_ID,
-	 * but we also want to check multiple services.  So instead, we
-	 * browse all rfcomm obex services.
-	 */
 	sdp_uuid16_create(&browse_uuid, PUBLIC_BROWSE_GROUP);
-	sdp_uuid16_create(&dun_id, DIALUP_NET_SVCLASS_ID);
+	sdp_uuid16_create(&service_id, svclass_id);
 	search = sdp_list_append (NULL, &browse_uuid);
-	search = sdp_list_append (search, &dun_id);
+	search = sdp_list_append (search, &service_id);
 
-	attrs = sdp_list_append (NULL, &proto_desc_list);
-	attrs = sdp_list_append (attrs, &svclass_id_list);
+	attrs = sdp_list_append (NULL, &range);
 
 	if (sdp_service_search_attr_req (sdp, search,
-					 SDP_ATTR_REQ_INDIVIDUAL, attrs,
+					 SDP_ATTR_REQ_RANGE, attrs,
 					 &recs))
 		goto end;
 
@@ -178,15 +203,21 @@ end:
 }
 
 int
-phonemgr_utils_get_channel (char *device)
+phonemgr_utils_get_channel (const char *device)
 {
 	bdaddr_t src, dst;
 	int channel;
 
+	/* If it's not a Bluetooth address */
+	if (bachk (device) < 0)
+		return -1;
+
 	bacpy (&src, BDADDR_ANY);
 	str2ba(device, &dst);
 
-	channel = find_dun_channel (&src, &dst);
+	channel = find_service_channel (&src, &dst, SERIAL_PORT_SVCLASS_ID);
+	if (channel < 0)
+		channel = find_service_channel (&src, &dst, DIALUP_NET_SVCLASS_ID);
 
 	return channel;
 }
@@ -194,12 +225,11 @@ phonemgr_utils_get_channel (char *device)
 char *
 phonemgr_utils_write_config (const char *driver, const char *addr, int channel)
 {
-	if (phonemgr_utils_is_bluetooth (addr) == FALSE) {
-		return g_strdup_printf ("[global]\n"
-			"port = %s\n"
-			"model = %s\n"
-			"connection = serial\n", addr, driver);
-	} else {
+	PhonemgrConnectionType type;
+
+	type = phonemgr_utils_address_is (addr);
+
+	if (type == PHONEMGR_CONNECTION_BLUETOOTH) {
 		if (channel > 0) {
 			return g_strdup_printf ("[global]\n"
 						"port = %s\n"
@@ -215,17 +245,39 @@ phonemgr_utils_write_config (const char *driver, const char *addr, int channel)
 						"model = %s\n"
 						"connection = bluetooth\n", addr, driver);
 		}
+	} else if (type == PHONEMGR_CONNECTION_USB) {
+		return g_strdup_printf ("[global]\n"
+					"port = %s\n"
+					"model = %s\n"
+					"connection = dku2libusb\n", addr, driver);
+	} else {
+		return g_strdup_printf ("[global]\n"
+			"port = %s\n"
+			"model = %s\n"
+			"connection = serial\n", addr, driver);
 	}
 }
 
+char *
+phonemgr_utils_config_append_debug (const char *config)
+{
+	return g_strdup_printf ("%s\n"
+				"[logging]\n"
+				"debug = on\n",
+				config);
+}
+
 static char *
 phonemgr_utils_driver_for_model (const char *model, const char *device)
 {
 	char *driver;
 
+	if (driver_model == NULL)
+		phonemgr_utils_init_hash_tables ();
+
 	driver = g_hash_table_lookup (driver_model, model);
 	if (driver == NULL) {
-		g_warning ("Model %s not supported natively", model);
+		g_message ("Model %s not supported natively", model);
 		driver = g_strdup (PHONEMGR_DEFAULT_DRIVER);
 	} else {
 		driver = g_strdup (driver);
@@ -233,7 +285,7 @@ phonemgr_utils_driver_for_model (const char *model, const char *device)
 
 	/* Add it to the list if it's a bluetooth device */
 	//FIXME this should also go in GConf
-	if (phonemgr_utils_is_bluetooth (device) != FALSE) {
+	if (phonemgr_utils_address_is (device) == PHONEMGR_CONNECTION_BLUETOOTH) {
 		g_hash_table_insert (driver_device,
 				g_strdup (device),
 				driver);
@@ -245,11 +297,18 @@ phonemgr_utils_driver_for_model (const char *model, const char *device)
 static char *
 phonemgr_utils_driver_for_device (const char *device)
 {
+	PhonemgrConnectionType type;
 	char *driver;
 
-	if (phonemgr_utils_is_bluetooth (device) == FALSE)
+	type = phonemgr_utils_address_is (device);
+	if (type == PHONEMGR_CONNECTION_USB)
+		return PHONEMGR_DEFAULT_USB_DRIVER;
+	if (type != PHONEMGR_CONNECTION_BLUETOOTH)
 		return NULL;
 
+	if (driver_device == NULL)
+		phonemgr_utils_init_hash_tables ();
+
 	driver = g_hash_table_lookup (driver_device, device);
 
 	return driver;
@@ -362,8 +421,6 @@ phonemgr_utils_guess_driver (PhonemgrState *phone_state, const char *device,
 	char *driver;
 	gn_error err;
 
-	phonemgr_utils_init_hash_tables ();
-
 	driver = phonemgr_utils_driver_for_device (device);
 	if (driver != NULL)
 		return driver;
@@ -389,24 +446,38 @@ bail:
 }
 
 PhonemgrState *
-phonemgr_utils_connect (const char *device, const char *driver, int channel, GError **error)
+phonemgr_utils_connect (const char *device, const char *driver, int channel, gboolean debug, GError **error)
 {
 	PhonemgrState *phone_state = NULL;
+	PhonemgrConnectionType type;
 	char *config, **lines;
+	const char *default_driver;
 	gn_data data;
 	struct gn_statemachine state;
 	gn_error err;
 
-	if (phonemgr_utils_is_bluetooth (device) != FALSE) {
-		if (phonemgr_utils_connection_is_supported (PHONEMGR_CONNECTION_BLUETOOTH) == FALSE) {
-			//FIXME set the error message
-			return NULL;
-		}
+	type = phonemgr_utils_address_is (device);
+	if (phonemgr_utils_connection_is_supported (type) == FALSE) {
+		//FIXME set the error message
+		return NULL;
 	}
+	if (type == PHONEMGR_CONNECTION_USB)
+		default_driver = PHONEMGR_DEFAULT_USB_DRIVER;
+	else
+		default_driver = PHONEMGR_DEFAULT_DRIVER;
 
-	config = phonemgr_utils_write_config (driver ? driver : PHONEMGR_DEFAULT_DRIVER, device, channel);
-	lines = g_strsplit (config, "\n", -1);
-	g_free (config);
+	config = phonemgr_utils_write_config (driver ? driver : default_driver, device, channel);
+	if (debug != FALSE) {
+		char *debug;
+
+		debug = phonemgr_utils_config_append_debug (config);
+		g_free (config);
+		lines = g_strsplit (debug, "\n", -1);
+		g_free (debug);
+	} else {
+		lines = g_strsplit (config, "\n", -1);
+		g_free (config);
+	}
 
 	if (gn_cfg_memory_read ((const char **)lines) < 0) {
 		g_warning ("gn_cfg_memory_read");
@@ -459,11 +530,27 @@ phonemgr_utils_tell_driver (const char *addr)
 {
 	GError *error = NULL;
 	PhonemgrState *phone_state;
+	PhonemgrConnectionType type;
 	const char *model;
+	char *driver;
 	int channel;
 
-	channel = phonemgr_utils_get_channel (addr);
-	phone_state = phonemgr_utils_connect (addr, NULL, channel, &error);
+	channel = -1;
+	type = phonemgr_utils_address_is (addr);
+	if (phonemgr_utils_connection_is_supported (type) == FALSE) {
+		g_warning ("Connection type isn't support by your gnokii build");
+		return;
+	}
+
+	if (type == PHONEMGR_CONNECTION_BLUETOOTH) {
+		channel = phonemgr_utils_get_channel (addr);
+		if (channel < 0) {
+			g_warning ("Couldn't find the channel to connect to on Bluetooth device");
+			return;
+		}
+	}
+
+	phone_state = phonemgr_utils_connect (addr, NULL, channel, FALSE, &error);
 	if (phone_state == NULL) {
 		g_warning ("Couldn't connect to the '%s' phone: %s", addr, PHONEMGR_CONDERR_STR(error));
 		if (error != NULL)
@@ -475,10 +562,73 @@ phonemgr_utils_tell_driver (const char *addr)
 
 	g_print ("model: '%s'\n", model);
 
+	if (type != PHONEMGR_CONNECTION_USB) {
+		driver = phonemgr_utils_driver_for_model (model, addr);
+		g_print ("guessed driver: '%s'\n", driver);
+		g_free (driver);
+	} else {
+		g_print ("guessed driver: '%s'\n", PHONEMGR_DEFAULT_USB_DRIVER);
+	}
+
 	phonemgr_utils_disconnect (phone_state);
 	phonemgr_utils_free (phone_state);
 }
 
+void
+phonemgr_utils_write_gnokii_config (const char *addr)
+{
+	GError *error = NULL;
+	PhonemgrConnectionType type;
+	char *driver, *config, *debug;
+	int channel;
+
+	channel = -1;
+	type = phonemgr_utils_address_is (addr);
+	if (phonemgr_utils_connection_is_supported (type) == FALSE) {
+		g_warning ("Connection type isn't support by your gnokii build");
+		return;
+	}
+
+	if (type == PHONEMGR_CONNECTION_BLUETOOTH) {
+		channel = phonemgr_utils_get_channel (addr);
+		if (channel < 0) {
+			g_warning ("Couldn't find the channel to connect to on Bluetooth device");
+			return;
+		}
+	}
+
+	if (type != PHONEMGR_CONNECTION_USB) {
+		PhonemgrState *phone_state;
+
+		phone_state = phonemgr_utils_connect (addr, NULL, channel, FALSE, &error);
+		if (phone_state == NULL) {
+			g_warning ("Couldn't connect to the '%s' phone: %s", addr, PHONEMGR_CONDERR_STR(error));
+			if (error != NULL)
+				g_error_free (error);
+			return;
+		}
+
+		driver = phonemgr_utils_guess_driver (phone_state, addr, NULL);
+
+		phonemgr_utils_disconnect (phone_state);
+		phonemgr_utils_free (phone_state);
+	} else {
+		driver = g_strdup (PHONEMGR_DEFAULT_USB_DRIVER);
+	}
+
+	config = phonemgr_utils_write_config (driver, addr, channel);
+	g_free (driver);
+	debug = phonemgr_utils_config_append_debug (config);
+	g_free (config);
+	if (g_file_set_contents ("gnokiirc", debug, -1, NULL) == FALSE) {
+		g_warning ("Couldn't write gnokiirc file in the current directory");
+	} else {
+		g_print ("Configuration file written in the current directory\n");
+		g_print ("Move gnokiirc to ~/.gnokiirc to start debugging with gnokii\n");
+	}
+	g_free (debug);
+}
+
 time_t
 gn_timestamp_to_gtime (gn_timestamp stamp)
 {
@@ -523,8 +673,6 @@ phonemgr_utils_connection_is_supported (PhonemgrConnectionType type)
 		break;
 	case PHONEMGR_CONNECTION_USB:
 		conntype = GN_CT_DKU2LIBUSB;
-		if (!gn_lib_is_connectiontype_supported (conntype))
-			conntype = GN_CT_DKU2;
 		break;
 	default:
 		g_assert_not_reached ();
